import { canJoin, liftTarget } from 'prosemirror-transform';
import { LIST_ITEM_NAME } from './const';
var getListItemType = function (doc) { return doc.type.schema.nodes.list_item; };
var isInList = function (state) {
    var itemType = getListItemType(state.doc);
    var _a = state.selection, $from = _a.$from, $to = _a.$to;
    if (!itemType || $from.parent.type !== itemType || $to.parent.type !== itemType)
        return false;
    return true;
};
var getListPos = function (state, from, to) {
    var itemType = getListItemType(state.doc);
    var listRes = [];
    state.doc.nodesBetween(from, to, function (node, pos) {
        if (node.type === itemType)
            listRes.push(pos + 1);
        return !node.isLeaf;
    });
    return listRes;
};
var checkAndMergeList = function (tr, pos) {
    var res = false;
    if (!isInList(tr))
        return res;
    var $pos = tr.doc.resolve(pos);
    var beforePos = pos;
    var afterPos = pos;
    if ($pos.depth) {
        beforePos = $pos.before();
        afterPos = $pos.after();
    }
    var $before = tr.doc.resolve(beforePos);
    var beforeNode = $before.nodeBefore;
    var beforeNodeAfter = $before.nodeAfter;
    var beforeNodePos = 0;
    if (beforeNode)
        beforeNodePos = beforePos - beforeNode.nodeSize;
    if (beforeNode && beforeNodeAfter && beforeNode.type === beforeNodeAfter.type && !beforeNode.type.spec.isolating && canJoin(tr.doc, beforePos)) {
        tr.join(beforePos);
        var nodeBefore = tr.doc.nodeAt(beforeNodePos);
        if (nodeBefore)
            afterPos = beforeNodePos + nodeBefore.nodeSize;
        res = true;
    }
    var $after = tr.doc.resolve(afterPos);
    var afterNode = $after.nodeBefore;
    var afterNodeAfter = $after.nodeAfter;
    if (afterNode && afterNodeAfter && afterNode.type === afterNodeAfter.type && !afterNode.type.spec.isolating && canJoin(tr.doc, afterPos)) {
        tr.join(afterPos);
        res = true;
    }
    return res;
};
// `start` after lift inherits the serial number of the previous list
var fixListStart = function (tr) {
    if (isInList(tr))
        return;
    var nodeBefore = tr.doc.resolve(tr.selection.$from.before()).nodeBefore;
    var afterPos = tr.selection.$from.after();
    var nodeAfter = tr.doc.resolve(afterPos).nodeAfter;
    if (!nodeBefore || !nodeAfter || nodeBefore.type !== nodeAfter.type)
        return;
    if (nodeBefore.type.spec.role !== 'list' || !nodeBefore.attrs.start)
        return;
    var start = parseInt(nodeBefore.attrs.start, 10);
    nodeBefore.content.forEach(function (child) {
        if (child.type !== nodeBefore.type)
            start++;
    });
    tr.setNodeMarkup(afterPos, undefined, { start: start });
};
var doLiftList = function (tr, pos) {
    var range = tr.doc.resolve(pos).blockRange();
    if (!range)
        return tr;
    if (range.depth) {
        var target = liftTarget(range);
        if (target === null || target === undefined)
            return tr;
        tr.lift(range, target ? target : range.depth - 1);
    }
    var $resPos = tr.doc.resolve(tr.mapping.map(range.start));
    if (!$resPos.depth || $resPos.parent.type.spec.role !== 'list') {
        if (!$resPos.nodeAfter)
            return tr;
        tr.setBlockType($resPos.pos, $resPos.pos + $resPos.nodeAfter.nodeSize, tr.doc.type.contentMatch.defaultType);
    }
    return tr;
};
var liftListItem = function (state, dispatch) {
    if (!isInList(state))
        return false;
    var tr = state.tr;
    var _a = state.selection, $from = _a.$from, $to = _a.$to;
    var listRes = getListPos(state, $from.pos, $to.pos).reverse();
    listRes.forEach(function (pos) { return doLiftList(tr, pos); });
    fixListStart(tr);
    dispatch && dispatch(tr);
    return true;
};
// at the start of the line, judged whether to clear the list format
var liftListItemAtHead = function (state, dispatch) {
    return state.selection.empty && !state.selection.$from.nodeBefore && liftListItem(state, dispatch);
};
var splitListItem = function (state, dispatch) {
    if (!dispatch || !isInList(state))
        return false;
    var tr = state.tr;
    var selection = state.selection, schema = state.schema;
    var $from = selection.$from, $to = selection.$to;
    var list_item = schema.nodes.list_item;
    if ($from.parent.content.size === 0) {
        liftListItem(state, dispatch);
        return true;
    }
    tr = tr.delete($from.pos, $to.pos);
    var attrs = $from.node().attrs;
    dispatch(tr
        .split($from.pos, 1, [
        {
            type: list_item,
            attrs: attrs,
        },
    ])
        .scrollIntoView());
    return true;
};
var splitListItemKeepMarks = function (state, dispatch) {
    return splitListItem(state, dispatch &&
        (function (tr) {
            var marks = state.storedMarks || (state.selection.$to.parentOffset && state.selection.$from.marks());
            if (marks)
                tr.ensureMarks(marks);
            dispatch(tr);
        }));
};
var checkOutMaxNestedLevel = function (state, maxLevel) {
    if (maxLevel === void 0) { maxLevel = 0; }
    var level = maxLevel;
    if (!level)
        return false;
    var $from = state.selection.$from;
    var depth = $from.depth - 1;
    while (level && depth) {
        var parent_1 = $from.node(depth--);
        if (parent_1.type.spec.role === 'list')
            level--;
    }
    return !level;
};
var sinkListItem = function (state, dispatch, canSink) {
    if (!dispatch || !isInList(state))
        return false;
    if (canSink && canSink())
        return true;
    var _a = state.selection, $from = _a.$from, $to = _a.$to;
    var type = state.schema.nodes.bullet_list || state.schema.nodes.ordered_list;
    var tr = state.tr;
    var listRes = getListPos(state, $from.pos, $to.pos);
    listRes.reverse().forEach(function (pos) {
        var $nFrom = tr.doc.resolve(pos);
        if ($nFrom.depth > 1)
            type = $nFrom.node(-1).type;
        if (!type)
            return;
        var range = $nFrom.blockRange();
        tr.wrap(range, [{ type: type }]);
        checkAndMergeList(tr, pos);
    });
    dispatch(tr.scrollIntoView());
    return true;
};
var filterKeymap = function (fn) { return function (editor, state, dispatch) { return fn(state, dispatch); }; };
var getListItem = function (slice) {
    var listNode = null;
    slice.content.descendants(function (node) {
        if (node.type.name === LIST_ITEM_NAME) {
            listNode = node;
            return false;
        }
        if (listNode)
            return false;
    });
    return listNode;
};
export { checkAndMergeList, checkOutMaxNestedLevel, filterKeymap, getListItem, liftListItem, liftListItemAtHead, sinkListItem, splitListItemKeepMarks, };
//# sourceMappingURL=utils.js.map